/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___IteratedFunctionSystems___
#define ___IteratedFunctionSystems___

#include <LibGenerator/ProceduralGenerator.h>
#include <map>

namespace ProceduralTextures
{
	struct IfsCell
	{
		UbPixel * m_pixel;
		Point<double, 2> m_ptPosition;

		IfsCell();
		IfsCell( const IfsCell & cell);
		virtual ~IfsCell();

		IfsCell & operator =( const IfsCell & cell);

		void Set( UbPixel * p_pixel, const Point<double, 2> & p_ptPosition);
	};

	typedef std::map	<	int,		IfsCell			>	IfsCellIntMap;
	typedef std::map	<	int,		IfsCellIntMap	>	IfsCellMap;
	typedef std::list	<	IfsCell						>	IfsCellLst;
	typedef std::vector <	IfsCell						>	IfsCellVct;

	class IteratedFunctionSystem : public ProceduralGenerator
	{
	public:
		typedef enum
		{	eIFS_GENERATOR_SIERPINSKI_CARPET
		,	eIFS_GENERATOR_SIERPINSKI_GASKET
		,	eIFS_GENERATOR_FRACTAL_LEAF
		,	eIFS_GENERATOR_HEIGHWAY_DRAGON
		,	eIFS_GENERATOR_MCWORTHER_PENTIGREE
		,	eIFS_GENERATOR_PENTADENTRITE
		,	eIFS_GENERATOR_KOCH_CURVE
		,	eIFS_GENERATOR_LEVY_DRAGON
		,	eIFS_GENERATOR_COUNT
		}	eIFS_GENERATOR;

	private:
		class Thread;

		class IfsGenerator
		{
		protected:
			int									m_iWidth;
			int									m_iHeight;
			PixelBuffer	*						m_pBackBuffer;
			size_t								m_uiVariationCount;
			int									m_iAngle;
			int									m_iShift;
			int									m_iSize;
			std::vector <Point<double, 3> >		m_arrayVariations;
			std::vector <Point<int, 2> >		m_arrayProbabilities;
			Point<double, 2>					m_ptPositionInSection;
			Point<int, 2>						m_ptMin;
			Point<int, 2>						m_ptMax;
			std::vector<IfsCellMap>				m_arrayInitCellsMap;
			std::vector<IfsCellLst>				m_arrayInitCellsLst;
			std::vector<IfsCellVct>				m_arrayInitCellsVct;
			UbPixel								m_pxColour;

		public:
			IfsGenerator( PixelBuffer * p_pBackBuffer, int p_iWidth, int p_iHeight, int p_iAngle, int p_iShift, int p_iSize, const Point<double, 2> & p_ptPosition, size_t p_uiVariationCount);
			virtual ~IfsGenerator();

			void Initialise( IfsCellMap & p_arrayGeneratedCellsMap, IfsCellLst & p_arrayGeneratedCellsLst, IfsCellVct & p_arrayGeneratedCellsVct);
			bool Generate( const IfsCell & p_cell, IfsCellMap & p_arrayGeneratedCells);
			bool Generate( const IfsCell & p_cell, IfsCellLst & p_arrayGeneratedCells);
			bool Generate( IfsCell p_cell, IfsCellVct & p_arrayGeneratedCells);

			bool NextGeneration();

			inline void SetAngle( int val) { m_iAngle = val;_initialiseVariations(); }
			inline void SetShift( int val) { m_iShift = val;_initialiseVariations(); }
			inline void SetSize( int val) { m_iSize = val;_initialiseVariations(); }

			inline int					GetShift			()const						{ return m_iShift; }
			inline int					GetAngle			()const						{ return m_iAngle; }
			inline int					GetSize				()const						{ return m_iSize; }
			inline size_t				GetVariationCount	()const						{ return m_uiVariationCount; }
			inline size_t				GetInitCellsMapSize	()const						{ return m_arrayInitCellsMap.size(); }
			inline IfsCellMap		&	GetInitCellsMap		( size_t p_uiIndex)			{ return m_arrayInitCellsMap[p_uiIndex]; }
			inline const IfsCellMap	&	GetInitCellsMap		( size_t p_uiIndex)const	{ return m_arrayInitCellsMap[p_uiIndex]; }
			inline size_t				GetInitCellsLstSize	()const						{ return m_arrayInitCellsLst.size(); }
			inline IfsCellLst		&	GetInitCellsLst		( size_t p_uiIndex)			{ return m_arrayInitCellsLst[p_uiIndex]; }
			inline const IfsCellLst	&	GetInitCellsLst		( size_t p_uiIndex)const	{ return m_arrayInitCellsLst[p_uiIndex]; }
			inline size_t				GetInitCellsVctSize	()const						{ return m_arrayInitCellsVct.size(); }
			inline IfsCellVct		&	GetInitCellsVct		( size_t p_uiIndex)			{ return m_arrayInitCellsVct[p_uiIndex]; }
			inline const IfsCellVct	&	GetInitCellsVct		( size_t p_uiIndex)const	{ return m_arrayInitCellsVct[p_uiIndex]; }

			void Resample();

		protected:
			bool _insertCell( const Point<double, 2> & p_ptPosition, IfsCellMap & p_arrayGeneratedCells, IfsCell * p_pResult = NULL);
			bool _insertCell( const Point<double, 2> & p_ptPosition, IfsCellLst & p_arrayGeneratedCells, IfsCell * p_pResult = NULL);
			bool _insertCell( const Point<double, 2> & p_ptPosition, IfsCellVct & p_arrayGeneratedCells, IfsCell * p_pResult = NULL);
			virtual void _initialiseVariations()=0;
		};

		class SierpinskiGasket : public IfsGenerator
		{
		public:
			SierpinskiGasket( PixelBuffer * p_pBackBuffer, int p_iWidth, int p_iHeight);

		private:
			virtual void _initialiseVariations();
		};

		class SierpinskiCarpet : public IfsGenerator
		{
		public:
			SierpinskiCarpet( PixelBuffer * p_pBackBuffer, int p_iWidth, int p_iHeight);

		private:
			virtual void _initialiseVariations();
		};

		class FractalLeaf : public IfsGenerator
		{
		public:
			FractalLeaf( PixelBuffer * p_pBackBuffer, int p_iWidth, int p_iHeight);

		private:
			virtual void _initialiseVariations();
		};

		class HeighwayDragon : public IfsGenerator
		{
		public:
			HeighwayDragon( PixelBuffer * p_pBackBuffer, int p_iWidth, int p_iHeight);

		private:
			virtual void _initialiseVariations();
		};

		class McWorterPentigree : public IfsGenerator
		{
		public:
			McWorterPentigree( PixelBuffer * p_pBackBuffer, int p_iWidth, int p_iHeight);

		private:
			virtual void _initialiseVariations();
		};

		class Pentadentrite : public IfsGenerator
		{
		public:
			Pentadentrite( PixelBuffer * p_pBackBuffer, int p_iWidth, int p_iHeight);

		private:
			virtual void _initialiseVariations();
		};

		class KochCurve : public IfsGenerator
		{
		public:
			KochCurve( PixelBuffer * p_pBackBuffer, int p_iWidth, int p_iHeight);

		private:
			virtual void _initialiseVariations();
		};

		class LevyDragon : public IfsGenerator
		{
		public:
			LevyDragon( PixelBuffer * p_pBackBuffer, int p_iWidth, int p_iHeight);

		private:
			virtual void _initialiseVariations();
		};

		class Thread : public ProceduralGenerator::Thread
		{
		private:

			IfsCellMap					*	m_pMapCells;
			IfsCellMap						m_mapGeneratedCells;
			IfsCellMap::iterator			m_itMapCells1;

			IfsCellLst					*	m_pLstCells;
			IfsCellLst						m_lstGeneratedCells;
			IfsCellLst::iterator			m_itLstCells1;

			IfsCellVct					*	m_pVctCells;
			size_t							m_uiCells1;
			size_t							m_uiCellsSize;

			PixelBuffer					*	m_backBuffer;
			IfsGenerator				**	m_pGenerator;
			bool							m_bResult;

		public:
			Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour);
			virtual ~Thread();

			virtual void Step();

			void Initialise( IfsCellMap * p_pMapCells, IfsCellLst * p_pLstCells, IfsCellVct * p_pVctCells, PixelBuffer * p_pBackBuffer, IfsGenerator ** p_pGenerator);

		private:
			void StepMap();
			void StepLst();
			void StepVct();
		};

	public:
		static wxString		ArrayNames[eIFS_GENERATOR_COUNT];

	private:
		IfsCellMap			m_mapCells;
		IfsCellLst			m_lstCells;
		IfsCellVct			m_VctCells;
		IfsGenerator *		m_pGenerator;
		IfsGenerator *		m_arrayGenerators[eIFS_GENERATOR_COUNT];

	public:
		IteratedFunctionSystem( wxEventType p_eventType, int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame);
		virtual ~IteratedFunctionSystem();

		virtual void SwapBuffers();

		int	GetShift	()const { return m_pGenerator->GetShift(); }
		int	GetAngle	()const { return m_pGenerator->GetAngle(); }
		int	GetSize		()const { return m_pGenerator->GetSize(); }

		void SetSize	( int val)  { m_pGenerator->SetSize( val); }
		void SetAngle	( int val)  { m_pGenerator->SetAngle( val); }
		void SetShift	( int val)  { m_pGenerator->SetShift( val); }

		void SetGenerator( eIFS_GENERATOR p_eGenerator);

	private:
		virtual void _glInit		(){}
		virtual void _glPreRender	(){}
		virtual void _glRender		( bool & p_bChanged){}
		virtual void _glPostRender	(){}
		virtual void _glCleanup		(){}

		void _initialise();
	};

	class IfsPanel : public ConfigPanel
	{
	private:
		typedef enum
		{	eAngle		= 50
		,	eShift		= 51
		,	eSize		= 52
		,	eGenerator	= 53
		} eIDs;

		IteratedFunctionSystem	*	m_pGenerator;
		wxTextCtrl				*	m_pEditAngle;
		wxTextCtrl				*	m_pEditShift;
		wxTextCtrl				*	m_pEditSize;
		wxComboBox				*	m_pComboGenerator;

	public:
		IfsPanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition=wxDefaultPosition, const wxSize & p_size=wxDefaultSize);
		virtual ~IfsPanel();

		void Create( IteratedFunctionSystem * p_pGenerator, wxWindow * p_pWindow, const wxPoint & p_ptPosition, const wxSize & p_size);

	private:
		DECLARE_EVENT_TABLE()
		void _onAngle			( wxCommandEvent	&	p_event);
		void _onShift			( wxCommandEvent	&	p_event);
		void _onSizeIfs			( wxCommandEvent	&	p_event);
		void _onGenerator		( wxCommandEvent	&	p_event);
	};

	class IfsPlugin : public Plugin<IfsPanel, IteratedFunctionSystem, IfsPlugin>
	{
	public:
#ifdef __WXMSW__
		IfsPlugin( wxEventType p_eventType=wxEVT_STATUS_UPDATE );
#else
		IfsPlugin( wxEventType p_eventType );
#endif
		virtual ~IfsPlugin();

		virtual wxString GetName();

		DECLARE_PLUGIN()
	};
}

#endif
