/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___ProceduralGenerator___
#define ___ProceduralGenerator___

#include "Prerequisites.h"

#include "Pixel.h"
#include "Thread.h"
#include <wx/wx.h>
#include <wx/tglbtn.h>
#include <map>
#include <list>
#include <vector>

#include "Plugin.h"
#include "GlObjects.h"

#if defined( _WIN32 )
#	define wxDEFAULT_HEIGHT	25
#else
#	define wxDEFAULT_HEIGHT	30
#endif

namespace ProceduralTextures
{
	typedef GlShaderProgram *	ShaderPool[10];

	class wxStatusUpdateEvent;

	typedef void ( * DrawFunc)( void * p_pThis);

	class ProceduralGenerator
	{
	public:
		typedef enum
		{	eEVENT_TYPE_SLIDER_UPDATE
		,	eEVENT_TYPE_THUMB_RELEASE
		,	eEVENT_TYPE_THUMB_TRACK
		,	eEVENT_TYPE_COMBO_SELECT
		,	eEVENT_TYPE_TEXT_ENTER
		,	eEVENT_TYPE_BUTTON_CLICKED
		,	eEVENT_TYPE_BUTTON_IMAGE
		,	eEVENT_TYPE_BUTTON_TOGGLE
		,	eEVENT_TYPE_COUNT
		}	eEVENT_TYPE;

	protected:
		class Thread : public wxThread
		{
		private:
			bool					m_bEnded;
			bool					m_bLaunched;
			bool					m_bStopped;

		protected:
			mutable wxMutex					m_mutex;
			ProceduralGenerator	*			m_pParent;
			int								m_iWidth;
			int								m_iTop;
			int								m_iBottom;
			int								m_iHeight;
			UbPixel							m_pxColour;
			size_t							m_uiIndex;

		public:
			Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour);
			virtual ~Thread();

			void Run();
			void Wait();
			virtual ExitCode Entry();

			virtual void Step()=0;

			inline bool	IsEnded		()const	{ return m_bEnded; }
			inline bool IsStopped	()const	{ m_mutex.Lock();bool l_bResult = m_bStopped;m_mutex.Unlock();return l_bResult; }

			inline void Relaunch	()						{ m_bLaunched = true; }
			inline void Stop		()						{ m_mutex.Lock();m_bStopped = true;m_mutex.Unlock(); }
			inline void SetRed		( unsigned char val)	{ m_pxColour.r = val; }
			inline void SetGreen	( unsigned char val)	{ m_pxColour.g = val; }
			inline void SetBlue		( unsigned char val)	{ m_pxColour.b = val; }
		};

	protected:
		wxSize						m_sizeDisplay;
		wxSize						m_sizeImage;
		unsigned long long			m_step;
		PixelBuffer					m_frontBuffer;
		PixelBuffer					m_backBuffer;
		UbPixel						m_pxColour;
		size_t						m_uiThreadCount;
		bool						m_bEnded;
		wxFrame *					m_pMainFrame;
		wxStatusUpdateEvent *		m_pEvent;
		bool						m_bNeedsConfigPanel;
		ConfigPanel *				m_pConfigPanel;
		std::vector< Thread * >		m_arraySlaveThreads;
		bool						m_initialised;
		GlVertexBuffer				m_vertexBuffer;
		GlIndexBuffer				m_indexBuffer;
		GlFrameBuffer				m_frameBuffer;
		GlTexture					m_texture;
		GlFrameBuffer				m_frameBufferSave;
		GlTexture					m_textureSave;
		bool						m_bToSave;
		wxEventType					m_eventType;

	public:
		ProceduralGenerator( wxEventType p_eventType, int p_width, int p_height, int p_iWndId, wxFrame * p_pMainFrame, bool p_bNeedsConfigPanel=true);
		virtual ~ProceduralGenerator();

		virtual bool Step();
		virtual void SetRed( unsigned char val);
		virtual void SetGreen( unsigned char val);
		virtual void SetBlue( unsigned char val);
		virtual void SwapBuffers();
		virtual void InitialiseStep();
		virtual void ClearAllThreads();
		virtual void GlInit();
		virtual void GlPreRender();
		virtual void GlRender( bool p_bChanged);
		virtual void GlPostRender();
		virtual void GlCleanup();

		bool AllEnded();
		void Suspend();
		void SetSize( int p_iWidth, int p_iHeight);
		void SetSize( const wxSize & p_size);
		void Resize( const wxSize & p_size);
		void SaveFrame();

		template <class ThreadClass>
		ThreadClass * CreateThread( int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour)
		{
			ThreadClass * l_pThread = new ThreadClass( this, m_arraySlaveThreads.size(), m_sizeImage.x, iTop, iBottom, iTotalHeight, p_pxColour);
			m_arraySlaveThreads.push_back( l_pThread);
			return l_pThread;
		}

		inline unsigned char		GetRed				()const { return m_pxColour.r; }
		inline unsigned char		GetGreen			()const { return m_pxColour.g; }
		inline unsigned char		GetBlue				()const { return m_pxColour.b; }
		inline bool					IsInitialised		()const { return m_initialised; }
		inline const UbPixel *		GetPixels			()const { return m_frontBuffer.const_ptr(); }
		inline const PixelBuffer *	GetPixelsBuffer		()const { return & m_frontBuffer; }
		inline PixelBuffer *		GetPixelsBuffer		()		{ return & m_frontBuffer; }
		inline wxSize const &		GetDisplaySize		()const { return m_sizeDisplay; }
		inline wxSize const &		GetImageSize		()const { return m_sizeImage; }
		inline bool					NeedsConfigPanel	()const	{ return m_bNeedsConfigPanel; }
		inline wxFrame *			GetMainFrame		()const	{ return m_pMainFrame; }
		inline ConfigPanel *		GetConfigPanel		()const	{ return m_pConfigPanel; }

		inline void	SetConfigPanel	( ConfigPanel * p_pPanel)			{ m_pConfigPanel = p_pPanel; }

	protected:
		virtual void _resize		( const wxSize & p_size);
		virtual void _glInit		()=0;
		virtual void _glPreRender	()=0;
		virtual void _glRender		( bool & p_bChanged)=0;
		virtual void _glPostRender	()=0;
		virtual void _glCleanup		()=0;
		void _cleanup();
		wxSize _loadImage( const wxString & p_strImagePath, wxImage & p_pImage, PixelBuffer & p_buffer);
		void _subRender();
		void _saveFrame();
	};

	class ConfigPanel
        :   public wxPanel
	{
    public:
	    ConfigPanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition=wxDefaultPosition, const wxSize & p_size=wxDefaultSize);
	    virtual ~ConfigPanel();
	};
}

#endif
