#include "ShaderWebcam/GeneratorShaderWebcam.h"

#include <wx/sysopt.h>
#include <wx/display.h>
#include <wx/tglbtn.h>
#include <cmath>
#include <wx/timer.h>

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************

Effect :: Effect( wxImage * p_pImage )
	:	m_pImage			( p_pImage )
	,	m_pShaderProgram	( NULL )
	,	m_strVertexFile		()
	,	m_strFragmentFile	()
	,	m_bNewShader		( false )
	,	m_llStartTime		( 0 )
	,	m_iWidth			( 0 )
	,	m_iHeight			( 0 )
{
	if( p_pImage && p_pImage->IsOk() )
	{
		m_iWidth	= p_pImage->GetWidth();
		m_iHeight	= p_pImage->GetHeight();
	}
}

Effect :: ~Effect()
{
	delete m_pShaderProgram;

	for (size_t i = 0 ; i < m_arrayImages.size() ; i++)
	{
		delete m_arrayImages[i];
	}

	m_arrayImages.clear();
	m_arrayImagesFiles.clear();
}

void Effect :: SetVertexFile( const wxString & p_strPath)
{
	m_strVertexFile = p_strPath;
}

void Effect :: SetFragmentFile( const wxString & p_strPath)
{
	m_strFragmentFile = p_strPath;
}

void Effect :: Compile()
{
	m_bNewShader = true;
}

void Effect :: SetImagePath( size_t p_iImage, const wxString & p_strImagePath)
{
	wxImage * l_pImage = NULL;

	if (p_iImage >= m_arrayImages.size())
	{
		l_pImage = new wxImage();
		m_arrayImages.push_back( l_pImage);
		m_arrayImagesFiles.push_back( p_strImagePath);
	}
	else
	{
		l_pImage = m_arrayImages[p_iImage];
		m_arrayImagesFiles[p_iImage] = p_strImagePath;
	}

	if (l_pImage->LoadFile( p_strImagePath))
	{
		int l_iWidth = l_pImage->GetWidth();
		int l_iHeight = l_pImage->GetHeight();
		std::cout << "Original : " << l_iWidth << "x" << l_iHeight << "\n";

		if (m_iWidth != l_iWidth || m_iHeight != l_iHeight)
		{
			std::cout << "Computed : " << m_iWidth << "x" << m_iHeight << "\n";
			l_pImage->Rescale( m_iWidth, m_iHeight, wxIMAGE_QUALITY_HIGH);
			std::cout << "Scaled : " << l_pImage->GetWidth() << "x" << l_pImage->GetHeight() << "\n";
		}
	}
	else
	{
		wxMessageBox( wxT( "No handler found for this image format"));
	}
}

void Effect :: GlPreRender()
{
	if (m_bNewShader)
	{
		if (m_pShaderProgram != NULL)
		{
			m_pShaderProgram->Cleanup();
		}
		else
		{
			m_pShaderProgram = new GlShaderProgram;
		}

		if ( ! m_strVertexFile.empty())
		{
			m_pShaderProgram->CreateObject( eSHADER_OBJECT_TYPE_VERTEX);
			m_pShaderProgram->SetProgramFile( m_strVertexFile, eSHADER_OBJECT_TYPE_VERTEX);
		}

		if ( ! m_strFragmentFile.empty())
		{
			m_pShaderProgram->CreateObject( eSHADER_OBJECT_TYPE_PIXEL);
			m_pShaderProgram->SetProgramFile( m_strFragmentFile, eSHADER_OBJECT_TYPE_PIXEL);
		}

		m_pShaderProgram->Initialise();

		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_width"),			0);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_height"),		0);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_time"),			0);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_texture"),		0);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_modif1"),		1);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_modif2"),		2);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_modif3"),		3);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_sep_type"),		0);
		m_pShaderProgram->CreateIntFrameVariable( wxT( "pg_sep_offset"),	0);

		m_llStartTime = wxGetLocalTimeMillis();
		m_bNewShader = false;
	}
}

void Effect :: GlPostRender()
{
}

void Effect :: GlCleanup()
{
	if (m_pShaderProgram != NULL)
	{
		m_pShaderProgram->Cleanup();
	}
}

bool Effect :: Activate( eSEPARATION p_eSeparationType, int p_iSeparationOffset)
{
	bool l_bReturn = true;

	if (m_pShaderProgram != NULL)
	{
		m_pShaderProgram->SetIntFrameVariableValue( wxT( "pg_width"),		m_iWidth);
		m_pShaderProgram->SetIntFrameVariableValue( wxT( "pg_height"),		m_iHeight);
		m_pShaderProgram->SetIntFrameVariableValue( wxT( "pg_sep_type"),	p_eSeparationType);
		m_pShaderProgram->SetIntFrameVariableValue( wxT( "pg_sep_offset"),	p_iSeparationOffset);
		m_pShaderProgram->SetIntFrameVariableValue( wxT( "pg_time"),		( wxGetLocalTimeMillis() - m_llStartTime).ToLong());
		m_pShaderProgram->Activate();
	}

	return l_bReturn;
}

void Effect :: Deactivate()
{
	if (m_pShaderProgram != NULL)
	{
		m_pShaderProgram->Deactivate();
	}
}

wxString Effect :: GetCompilerLog( eSHADER_OBJECT_TYPE p_eType)
{
	wxString l_strReturn;

	if (m_pShaderProgram != NULL)
	{
		GlShaderObject * l_pShader = m_pShaderProgram->GetProgram( p_eType);

		if (l_pShader != NULL)
		{
			l_strReturn = l_pShader->GetCompilerLog();
		}
	}

	return l_strReturn;
}

wxString Effect :: GetLinkerLog()
{
	wxString l_strReturn;

	if (m_pShaderProgram != NULL)
	{
		l_strReturn = m_pShaderProgram->GetLinkerLog();
	}

	return l_strReturn;
}

void Effect :: UpdateImages()
{
	m_iWidth = m_pImage->GetWidth();
	m_iHeight = m_pImage->GetHeight();

	for (size_t i = 0 ; i < m_arrayImages.size() ; i++)
	{
		m_arrayImages[i]->LoadFile( m_arrayImagesFiles[i]);
		m_arrayImages[i]->ResampleBicubic( m_iWidth, m_iHeight);
	}
}

//*************************************************************************************************

ShaderWebcam :: ShaderWebcam( wxEventType p_eventType, int p_width, int p_height, int p_iWndId, wxFrame * p_pFrame)
	:	ProceduralGenerator( p_eventType, p_width, p_height, p_iWndId, p_pFrame, true)
	,	m_pCapture			( NULL )
	,	m_iSeparationOffset	( 50)
	,	m_eSeparationType	( eSEPARATION_NONE)
{
	m_pCapture = cvCaptureFromCAM( CV_CAP_ANY );

	if( m_pCapture )
	{
		m_sizeImage.x = cvGetCaptureProperty( m_pCapture, CV_CAP_PROP_FRAME_WIDTH );
		m_sizeImage.y = cvGetCaptureProperty( m_pCapture, CV_CAP_PROP_FRAME_HEIGHT );
		m_frontBuffer.init( Point< size_t, 2 >( m_sizeImage.x, m_sizeImage.y ) );
		m_backBuffer.init( Point< size_t, 2 >( m_sizeImage.x, m_sizeImage.y ) );
	}

	_resize( m_sizeImage );

	m_frameBuffer1.BindTexture( GL_COLOR_ATTACHMENT0, & m_texture1 );
	m_frameBuffer2.BindTexture( GL_COLOR_ATTACHMENT0, & m_texture2 );
	wxSystemOptions::SetOption( wxT( "msw.remap" ), 0 );
	wxInitAllImageHandlers();

	m_initialised = m_pCapture != NULL;

	if( m_initialised )
	{
		m_initialised = EffectAdd();
	}
}

ShaderWebcam :: ~ShaderWebcam()
{
	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		delete m_arrayEffects[i];
	}

	cvReleaseCapture( &m_pCapture );
	m_arrayEffects.clear();
	wxImage::CleanUpHandlers();
	m_initialised = false;
	_cleanup();
}

void ShaderWebcam :: SetImagePath( const wxString & p_strImagePath)
{
	m_initialised = false;
	wxSize l_size = _loadImage( p_strImagePath, m_image, m_backBuffer);
	m_frontBuffer.init( Point<size_t, 2>( l_size.x, l_size.y));
	SetSize( l_size);

	if (m_arrayEffects.size() == 0)
	{
		EffectAdd();
	}
	else
	{
		for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
		{
			m_arrayEffects[i]->UpdateImages();
		}
	}

	m_initialised = true;
}

void ShaderWebcam :: ResetTime()
{
	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		m_arrayEffects[i]->ResetTime();
	}
}

wxSize ShaderWebcam :: _loadImage( PixelBuffer & p_buffer )
{
	m_pImage = cvQueryFrame( m_pCapture );
	wxSize l_sizeReturn;

	if( m_pImage )
	{
		l_sizeReturn.x = m_pImage->width;
		l_sizeReturn.y = m_pImage->height;

		p_buffer.set< byte, 3, 0, true, false >( (byte *)m_pImage->imageData, NULL );
	}

	return l_sizeReturn;
}

void ShaderWebcam :: _glInit()
{
	m_texture1.Initialise();
	m_texture2.Initialise();

	m_frameBuffer1.Initialise();
	m_frameBuffer2.Initialise();
}

void ShaderWebcam :: _glPreRender()
{
	_loadImage( m_backBuffer );
	m_texture1.Initialise( & m_backBuffer);
	m_texture2.Initialise( m_sizeImage);

	m_frameBuffer1.Initialise( this);
	m_frameBuffer2.Initialise( this);

	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		m_arrayEffects[i]->GlPreRender();
	}

	OpenGl::Viewport( 0, 0, m_sizeImage.x, m_sizeImage.y);
}

void ShaderWebcam :: _glRender( bool & p_bChanged)
{
	GlFrameBuffer * l_pFrameBufferTexture = & m_frameBuffer1;
	GlFrameBuffer * l_pFrameBufferRender = & m_frameBuffer2;

	m_texture1.Activate();
	m_texture1.UploadGlImage();
	m_texture1.Deactivate();

	OpenGl::MatrixMode( GL_PROJECTION);
	OpenGl::LoadIdentity();
	OpenGl::Ortho( 0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
	OpenGl::MatrixMode( GL_MODELVIEW);
	OpenGl::LoadIdentity();

	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		if (l_pFrameBufferRender->Activate())
		{
			if (EffectActivate( i))
			{
				OpenGl::Clear( GL_COLOR_BUFFER_BIT);
				l_pFrameBufferTexture->GetTexture( GL_COLOR_ATTACHMENT0)->Activate();
				_subRender();
				l_pFrameBufferTexture->GetTexture( GL_COLOR_ATTACHMENT0)->Deactivate();
				EffectDeactivate( i);
			}

			l_pFrameBufferRender->Deactivate();
		}

		std::swap( l_pFrameBufferTexture, l_pFrameBufferRender);
	}

	OpenGl::MatrixMode( GL_PROJECTION);
	OpenGl::LoadIdentity();
	OpenGl::Ortho( 0.0, 1.0, 1.0, 0.0, 0.0, 1.0);
	OpenGl::MatrixMode( GL_MODELVIEW);
	OpenGl::LoadIdentity();

	m_frameBuffer.Activate();
	OpenGl::Clear( GL_COLOR_BUFFER_BIT);
	l_pFrameBufferTexture->GetTexture( GL_COLOR_ATTACHMENT0)->Activate();
	_subRender();
	l_pFrameBufferTexture->GetTexture( GL_COLOR_ATTACHMENT0)->Deactivate();
	m_frameBuffer.Deactivate();
}

void ShaderWebcam :: _glPostRender()
{
	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		m_arrayEffects[i]->GlPostRender();
	}
}

void ShaderWebcam :: _glCleanup()
{
	for (size_t i = 0 ; i < m_arrayEffects.size() ; i++)
	{
		m_arrayEffects[i]->GlCleanup();
		delete m_arrayEffects[i];
	}

	m_arrayEffects.clear();
}

bool ShaderWebcam :: EffectAdd()
{
	Effect * l_pEffect = new Effect( & m_image );
	m_arrayEffects.push_back( l_pEffect );
	return true;
}

bool ShaderWebcam :: EffectRemove( size_t p_uiIndex)
{
	bool l_bReturn = p_uiIndex < m_arrayEffects.size();

	if (l_bReturn)
	{
		Effect * l_pEffect = m_arrayEffects[p_uiIndex];
		std::vector <Effect *>::iterator l_it = m_arrayEffects.begin();
		m_arrayEffects.erase( l_it + p_uiIndex);
		delete l_pEffect;
	}

	return l_bReturn;
}

void ShaderWebcam :: EffectSetVertexFile( size_t p_uiIndex, const wxString & p_strPath)
{
	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->SetVertexFile( p_strPath);
	}
}

void ShaderWebcam :: EffectSetFragmentFile( size_t p_uiIndex, const wxString & p_strPath)
{
	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->SetFragmentFile( p_strPath);
	}
}

void ShaderWebcam :: EffectSetImagePath( size_t p_uiIndex, size_t p_iImage, const wxString & p_strImagePath)
{
	m_initialised = false;

	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->SetImagePath( p_iImage, p_strImagePath);
	}

	m_initialised = true;
}

void ShaderWebcam :: EffectCompile( size_t p_uiIndex)
{
	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->Compile();
	}
}

wxString ShaderWebcam :: EffectGetCompilerLog( size_t p_uiIndex, eSHADER_OBJECT_TYPE p_eType)
{
	wxString l_strReturn;

	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->GetCompilerLog( p_eType);
	}

	return l_strReturn;
}

wxString ShaderWebcam :: EffectGetLinkerLog( size_t p_uiIndex)
{
	wxString l_strReturn;

	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->GetLinkerLog();
	}

	return l_strReturn;
}

bool ShaderWebcam :: EffectActivate( size_t p_uiIndex)
{
	bool l_bReturn = p_uiIndex < m_arrayEffects.size();

	if (l_bReturn)
	{
		l_bReturn = m_arrayEffects[p_uiIndex]->Activate( m_eSeparationType, m_iSeparationOffset);
	}

	return l_bReturn;
}

void ShaderWebcam :: EffectDeactivate( size_t p_uiIndex)
{
	if (p_uiIndex < m_arrayEffects.size())
	{
		m_arrayEffects[p_uiIndex]->Deactivate();
	}
}

void ShaderWebcam :: _resize( const wxSize & p_size)
{
	ProceduralGenerator::_resize( p_size);
	m_frameBuffer1.Resize( m_sizeImage);
	m_frameBuffer2.Resize( m_sizeImage);
	m_texture1.Resize( m_sizeImage);
	m_texture2.Resize( m_sizeImage);
}

//******************************************************************************

IMPLEMENT_PLUGIN( SwPlugin, SwPanel, ShaderWebcam)

SwPlugin :: SwPlugin( wxEventType p_eventType )
	:	Plugin< SwPanel, ShaderWebcam, SwPlugin >( p_eventType )
{
}

SwPlugin :: ~SwPlugin()
{
}

wxString SwPlugin :: GetName()
{
#if defined( NDEBUG )
	return wxT( "Shader Webcam");
#else
	return wxT( "Shader Webcam (Debug)");
#endif
}

//*************************************************************************************************

SwPanel :: SwPanel( wxWindow * p_pWindow, wxWindowID p_id, const wxPoint & p_ptPosition, const wxSize & p_size)
	:	ConfigPanel( p_pWindow, p_id, p_ptPosition, p_size)
	,	m_pShadersConfigs( NULL)
{
}

SwPanel :: ~SwPanel()
{
}

void SwPanel :: Create( ShaderWebcam * p_pGenerator, wxWindow * p_pParent, const wxPoint & p_ptPosition, const wxSize & p_size)
{
	SetSize( p_ptPosition.x, p_ptPosition.y, p_size.x, p_size.y);
	Reparent( p_pParent);
	m_pGenerator = p_pGenerator;

	m_pButtonResetTime = new wxButton( this, eResetTime, wxT( "Reset Time"), wxPoint( 20, 35), wxSize( p_size.x - 40, wxDEFAULT_HEIGHT));
//	m_pButtonResetTime->Hide();

	wxArrayString l_arrayChoices;
	l_arrayChoices.push_back( wxT( "Not separated"));
	l_arrayChoices.push_back( wxT( "Horizontally"));
	l_arrayChoices.push_back( wxT( "Vertically"));
	m_pComboSeparationType = new wxComboBox( this, eSepType, wxT( "Not separated"), wxPoint( 20, 70), wxSize( p_size.x - 40, wxDEFAULT_HEIGHT), l_arrayChoices, wxCB_READONLY);
	m_pComboSeparationType->SetFont( GetFont());
//	m_pComboSeparationType->Hide();

	m_pSliderSeparationOffset = new wxSlider( this, eSepOffset, 50, 0, 100, wxPoint( 20, 100), wxSize( 100, 20));
//	m_pSliderSeparationOffset->Hide();

	m_pShadersConfigs = new SwEffectsNotebook( this, eShaders, wxPoint( 0, 125), wxSize( p_size.x, p_size.y - 135));
	m_pShadersConfigs->CreatePage( m_pGenerator, wxT( "Shader 1"));
	m_pShadersConfigs->CreatePage( m_pGenerator, wxT( "New ..."));
	m_pShadersConfigs->Layout();
//	m_pShadersConfigs->Hide();
}

BEGIN_EVENT_TABLE( SwPanel, ConfigPanel)
	EVT_SIZE(							SwPanel::_onSize)
	EVT_COMBOBOX(		eSepType,		SwPanel::_onSepType)
	EVT_SLIDER(			eSepOffset,		SwPanel::_onSepOffset)
	EVT_BUTTON(			eResetTime,		SwPanel::_onResetTime)
END_EVENT_TABLE()

void SwPanel :: _onSize( wxSizeEvent & p_event)
{
	if (m_pShadersConfigs != NULL)
	{
		m_pShadersConfigs->SetSize( p_event.GetSize().x, p_event.GetSize().y - 135);
	}

	p_event.Skip();
}

void SwPanel :: _onSepType( wxCommandEvent	& p_event)
{
	m_pGenerator->SetSeparationType( eSEPARATION( p_event.GetInt()));

	if (p_event.GetInt() != 0)
	{
		m_pSliderSeparationOffset->Show();
	}
	else
	{
		m_pSliderSeparationOffset->Hide();
	}
}

void SwPanel :: _onSepOffset( wxCommandEvent & p_event)
{
	m_pGenerator->SetSeparationOffset( p_event.GetInt());
}

void SwPanel :: _onResetTime( wxCommandEvent & p_event)
{
	m_pGenerator->ResetTime();
}

//*************************************************************************************************

SwEffectPage :: SwEffectPage( SwEffectsNotebook * p_pParent, ShaderWebcam * p_pGenerator, size_t p_uiIndex, bool p_bCanRemove, const wxPoint & p_ptPosition, const wxSize & p_size)
	:	wxNotebookPage( p_pParent, wxID_ANY, p_ptPosition, p_size, wxBORDER_NONE | wxCLIP_CHILDREN)
	,	m_pParent( p_pParent)
	,	m_uiIndex( p_uiIndex)
	,	m_pGenerator( p_pGenerator)
{
	m_pButtonVertexFile		= new wxToggleButton(	this,	eVertexFile,	wxT( "Vertex shader"),		wxPoint( 0, 0),		wxSize( p_size.x, wxDEFAULT_HEIGHT));
	m_pButtonFragmentFile	= new wxToggleButton(	this,	eFragmentFile,	wxT( "Fragment shader"),	wxPoint( 0, 35),	wxSize( p_size.x, wxDEFAULT_HEIGHT));
	m_pButtonCompile		= new wxButton(			this,	eCompile,		wxT( "Compile"),			wxPoint( 0, 70),	wxSize( p_size.x, wxDEFAULT_HEIGHT));
	m_pButtonCompilerLog	= new wxButton(			this,	eCompilerLog,	wxT( "Compile result"),		wxPoint( 0, 105),	wxSize( p_size.x, wxDEFAULT_HEIGHT));

	if (p_bCanRemove)
	{
		m_pButtonRemove		= new wxButton(			this,	eRemove,		wxT( "Remove"),				wxPoint( 0, 140),	wxSize( p_size.x, wxDEFAULT_HEIGHT));
	}

	m_pButtonCompile->Disable();
	m_pButtonCompilerLog->Disable();
}

SwEffectPage :: ~SwEffectPage()
{
}

BEGIN_EVENT_TABLE( SwEffectPage, wxNotebookPage)
	EVT_TOGGLEBUTTON(	eVertexFile,	SwEffectPage::_onVertexShaderPath)
	EVT_TOGGLEBUTTON(	eFragmentFile,	SwEffectPage::_onFragmentShaderPath)
	EVT_BUTTON(			eCompile,		SwEffectPage::_onShaderCompile)
	EVT_BUTTON(			eCompilerLog,	SwEffectPage::_onCompilerLog)
	EVT_BUTTON(			eRemove,		SwEffectPage::_onRemove)
END_EVENT_TABLE()

void SwEffectPage :: _onVertexShaderPath( wxCommandEvent & p_event)
{
	wxString l_strFile;

	if (p_event.IsChecked())
	{
		wxString l_strWildcard = wxT( "Tous Fichiers (*.*)|*.*");
		l_strFile = wxFileSelector( wxT( "Choose a file"), wxT( "./data"), wxEmptyString, wxEmptyString, l_strWildcard);
	}

	m_pButtonCompile->Enable();
	m_pButtonCompilerLog->Disable();
	m_pGenerator->EffectSetVertexFile( m_uiIndex, l_strFile);
}

void SwEffectPage :: _onFragmentShaderPath( wxCommandEvent & p_event)
{
	wxString l_strFile;

	if (p_event.IsChecked())
	{
		wxString l_strWildcard = wxT( "Tous Fichiers (*.*)|*.*");
		l_strFile = wxFileSelector( wxT( "Choose a file"), wxT( "./data"), wxEmptyString, wxEmptyString, l_strWildcard);
	}

	m_pButtonCompile->Enable();
	m_pButtonCompilerLog->Disable();
	m_pGenerator->EffectSetFragmentFile( m_uiIndex, l_strFile);
}

void SwEffectPage :: _onShaderCompile( wxCommandEvent & p_event)
{
	m_pGenerator->EffectCompile( m_uiIndex);
	m_pButtonCompile->Disable();
	m_pButtonCompilerLog->Enable();
}

void SwEffectPage :: _onCompilerLog( wxCommandEvent & p_event)
{
	wxString l_strVertexLog = m_pGenerator->EffectGetCompilerLog( m_uiIndex, eSHADER_OBJECT_TYPE_VERTEX);
	wxString l_strFragmentLog = m_pGenerator->EffectGetCompilerLog( m_uiIndex, eSHADER_OBJECT_TYPE_PIXEL);
	wxString l_strLinkerLog = m_pGenerator->EffectGetLinkerLog( m_uiIndex);
	wxString l_strText;

	if ( ! l_strVertexLog.empty())
	{
		l_strText += wxT( "Vertex Shader :\n") + l_strVertexLog;
	}

	if ( ! l_strFragmentLog.empty())
	{
		if ( ! l_strText.empty())
		{
			l_strText += wxT( "\n\n");
		}

		l_strText += wxT( "Fragment Shader :\n") + l_strFragmentLog + wxT( "\n\n");
	}

	if ( ! l_strLinkerLog.empty())
	{
		if ( ! l_strText.empty())
		{
			l_strText += wxT( "\n\n");
		}

		l_strText += wxT( "Linker :\n") + l_strLinkerLog + wxT( "\n\n");
	}

	if (l_strText.empty())
	{
		l_strText = wxT( "No error");
	}

	wxMessageBox( l_strText, wxT( "Compile and Link result"));
}

void SwEffectPage :: _onRemove( wxCommandEvent & p_event)
{
	if (m_pGenerator->EffectRemove( m_uiIndex))
	{
		m_pParent->ChangeSelection( m_uiIndex - 1);
		m_pParent->DeletePage( m_uiIndex);
	}
}

//*************************************************************************************************

SwEffectsNotebook :: SwEffectsNotebook( wxWindow * p_pParent, wxWindowID p_id, const wxPoint & p_ptPosition, const wxSize & p_size, long lStyle)
	:	wxNotebook( p_pParent, p_id, p_ptPosition, p_size, lStyle)
	,	m_pSelection( NULL)
	,	m_pGenerator( NULL)
{
	SetBackgroundColour( p_pParent->GetBackgroundColour());
	SetForegroundColour( p_pParent->GetForegroundColour());
}

SwEffectsNotebook :: ~SwEffectsNotebook()
{
}

SwEffectPage * SwEffectsNotebook :: CreatePage( ShaderWebcam * p_pGenerator, const wxString & p_strTitle)
{
	wxSize l_size = GetClientSize();
	SwEffectPage * l_pReturn = new SwEffectPage( this, p_pGenerator, m_listPages.size(), m_listPages.size() > 0, wxPoint( 10, 30), l_size + wxSize( -20, -40));
	m_listPages.push_back( l_pReturn);
	AddPage( l_pReturn, p_strTitle);

	if (m_pSelection == NULL)
	{
		m_pSelection = l_pReturn;
	}

	if (m_pGenerator == NULL)
	{
		m_pGenerator = p_pGenerator;
	}

	return l_pReturn;
}

bool SwEffectsNotebook :: DeletePage( size_t p_uiPage)
{
	bool l_bReturn = wxNotebook::DeletePage( p_uiPage);

	if (l_bReturn)
	{
		std::vector <SwEffectPage *>::iterator l_it = m_listPages.begin();
		m_listPages.erase( l_it + p_uiPage);
		l_it = l_it + p_uiPage;

		while (l_it != m_listPages.end())
		{
			( * l_it)->DecIndex();
			++l_it;
		}
	}

	return l_bReturn;
}

BEGIN_EVENT_TABLE( SwEffectsNotebook, wxNotebook)
	EVT_SIZE(										SwEffectsNotebook::_onSize)
	EVT_NOTEBOOK_PAGE_CHANGED( SwPanel::eShaders,	SwEffectsNotebook::_onSelectPage)
END_EVENT_TABLE()

void SwEffectsNotebook :: _onSize( wxSizeEvent & p_event)
{
	wxPoint l_ptPosition = wxPoint( 10, 30);
	wxSize l_size = GetClientSize() + wxSize( -20, -40);
	for (std::vector<SwEffectPage *>::iterator l_it = m_listPages.begin() ; l_it != m_listPages.end() ; ++l_it)
	{
		( * l_it)->SetSize( l_ptPosition.x, l_ptPosition.y, l_size.x, l_size.y);
	}
}

void SwEffectsNotebook :: _onSelectPage( wxBookCtrlEvent & p_event)
{
	int l_iSelection = p_event.GetSelection();

	if (l_iSelection == m_listPages.size() - 1 && m_pGenerator && m_pGenerator->EffectAdd())
	{
		wxString l_strPage = wxT( "Shader ");
		l_strPage << m_listPages.size();
		SetPageText( l_iSelection, l_strPage);
		CreatePage( m_pGenerator, wxT( "New ..."));
	}
}

//*************************************************************************************************
